/*
    Instruction:
    1. supply the sorter with external supply 12V
    2. connect USB to Arduino, configure COM port
    3. upload the code to the Arduino
    4. while pressing the Switch, run Putty at your COM port, 115200 bods
    5. keep Switch pressed until you see message:


*/

#include <Servo.h>
#include <math.h>
#include <Wire.h>
#include <Adafruit_TCS34725.h>
#include <EEPROM.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>

/****************************************/
#define N_VESSELS 9
#define TARGET_PERIOD_MS 930.0

/*averaged values of white beads for your beads sorter {R,G,B}*/
const float BS_wav[] PROGMEM = 
{
    1283.57, 1201.17, 647.12
};

/****************************************/

#define EEADR_CALIBR_DONE 0  // EEPROM address for 1 char to check if Calibration was done or not.
    #define CALIBRATION_DONE_CHAR 0xAB  
#define EEADR_MOTOR_PWM 1    // EEPROM address for motor PWM (1 Byte)
#define EEADR_SERVO_BASE 2   // EEPROM start address for servo angles array (9 Bytes) 

#define THRASH_VESSEL_IDX 0     //index of vessel for beads not currently sorted or NONES

#define M_FORWARD 1
#define M_BACKWARD 0

#define motor_speed(s) (OCR2A = s)
#define PIN_OFF(port, pin) (port &= ~(1<<pin))
#define PIN_ON(port, pin) (port |= (1<<pin))

/*labels*/
const char label_0[] PROGMEM = "none/black ";
const char label_1[] PROGMEM = "Pink       ";
const char label_2[] PROGMEM = "Red        ";
const char label_3[] PROGMEM = "White      ";
const char label_4[] PROGMEM = "Green      ";
const char label_5[] PROGMEM = "Blue       ";
const char label_6[] PROGMEM = "Yellow     ";
const char label_7[] PROGMEM = "Lt Green   ";
const char label_8[] PROGMEM = "Violet     ";
const char label_9[] PROGMEM = "Lt Violet  ";
const char label_10[] PROGMEM = "Orange    ";
const char label_11[] PROGMEM = "Lt Orange ";
const char label_12[] PROGMEM = "Grey      ";
const char label_13[] PROGMEM = "Creme     ";
const char label_14[] PROGMEM = "Blue Sky  ";
const char label_15[] PROGMEM = "Lt Pink   ";
const char label_16[] PROGMEM = "Lt Yellow ";
const char label_17[] PROGMEM = "Brown     ";

const char *const labelTable[] PROGMEM = {
    label_0, label_1, label_2, label_3, label_4, label_5,
    label_6, label_7, label_8, label_9, label_10, label_11,
    label_12, label_13, label_14, label_15, label_16, label_17};


/******************************************************************
        Start of NN defines, variables and data
        Arduino resources usage:
        > 812 Bytes of PROGMEM space for NN weights and parameters
        > 44 Bytes of RAM space for variables
 ******************************************************************/

typedef uint8_t uint8;

#define N_INPUT_LAYERS 3
#define N_HIDDEN_LAYERS 8
#define N_LABELS 18

float __nn_a2[N_HIDDEN_LAYERS + 1]; /*+1 to include bias term*/ 
float __nn_X[N_INPUT_LAYERS + 1];  /*+1 to include bias term*/ 

const float NN_wav[] PROGMEM = {
3178.58297872,	3004.63191489,	1621.39574468
};

const float NN_mu[] PROGMEM = {
1502.90346832,	1291.40250388,	1272.73718545
};

const float NN_sigma[] PROGMEM = {
780.78902462,	688.36456605,	665.86560294
};

const uint8 NN_thetas[] PROGMEM = {
29,	145,	222,	188,	23,	87,	102,	192,	157,	100,	
150,	64,	61,	137,	164,	62,	229,	214,	114,	192,	
16,	86,	185,	61,	136,	15,	79,	65,	154,	48,	
130,	193,	233,	152,	54,	63,	102,	36,	155,	64,	
116,	226,	47,	193,	4,	150,	158,	64,	180,	244,	
138,	193,	23,	124,	48,	194,	249,	165,	31,	66,	
92,	201,	216,	192,	61,	196,	33,	63,	254,	195,	
239,	63,	147,	63,	139,	191,	51,	90,	73,	64,	
228,	214,	244,	192,	210,	181,	125,	63,	33,	71,	
184,	193,	63,	17,	117,	65,	38,	78,	39,	64,	
245,	155,	19,	65,	136,	238,	1,	65,	178,	33,	
100,	193,	226,	133,	216,	192,	118,	17,	139,	193,	
13,	49,	19,	65,	246,	219,	35,	65,	174,	115,	
163,	193,	227,	125,	108,	193,	133,	229,	253,	192,	
39,	217,	131,	65,	70,	15,	193,	65,	70,	21,	
19,	193,	65,	242,	246,	63,	202,	151,	2,	193,	
191,	177,	181,	193,	155,	145,	55,	193,	156,	175,	
203,	193,	233,	5,	193,	193,	237,	145,	113,	191,	
105,	154,	9,	193,	215,	102,	77,	65,	155,	189,	
24,	65,	197,	212,	204,	190,	164,	226,	20,	193,	
174,	179,	74,	193,	201,	73,	71,	193,	206,	129,	
84,	193,	113,	33,	143,	192,	227,	8,	109,	193,	
200,	201,	231,	193,	61,	156,	132,	65,	94,	73,	
146,	65,	65,	166,	202,	192,	161,	70,	235,	64,	
52,	119,	25,	193,	52,	170,	146,	193,	93,	95,	
53,	193,	143,	221,	5,	193,	122,	93,	47,	193,	
236,	119,	43,	193,	121,	109,	235,	64,	4,	161,	
122,	65,	209,	240,	201,	191,	182,	221,	13,	62,	
213,	71,	97,	65,	58,	170,	202,	193,	119,	47,	
12,	65,	45,	114,	129,	193,	247,	240,	27,	193,	
182,	134,	203,	193,	107,	56,	151,	63,	4,	145,	
173,	193,	152,	145,	202,	192,	86,	42,	87,	193,	
35,	191,	131,	63,	73,	170,	51,	65,	70,	12,	
161,	192,	134,	54,	64,	65,	100,	219,	239,	192,	
47,	118,	79,	65,	61,	15,	65,	64,	45,	20,	
217,	193,	217,	170,	67,	65,	59,	192,	24,	194,	
218,	96,	87,	193,	58,	94,	2,	193,	158,	158,	
217,	192,	66,	224,	35,	65,	43,	110,	154,	192,	
49,	57,	225,	193,	48,	154,	210,	191,	251,	16,	
246,	61,	75,	82,	77,	193,	103,	22,	127,	65,	
153,	24,	203,	191,	65,	37,	11,	193,	134,	204,	
178,	65,	24,	84,	167,	192,	143,	251,	156,	193,	
49,	214,	155,	65,	165,	182,	92,	193,	98,	107,	
225,	191,	245,	61,	170,	193,	85,	43,	19,	63,	
39,	219,	176,	65,	57,	32,	171,	193,	221,	233,	
143,	192,	31,	249,	120,	63,	126,	49,	41,	192,	
180,	27,	85,	193,	101,	150,	36,	65,	71,	67,	
138,	192,	144,	42,	232,	192,	122,	84,	104,	193,	
197,	57,	250,	193,	165,	212,	210,	64,	26,	83,	
140,	193,	238,	12,	18,	194,	53,	93,	7,	63,	
122,	215,	216,	65,	218,	50,	165,	192,	219,	97,	
111,	193,	162,	22,	133,	193,	3,	27,	237,	63,	
102,	230,	163,	192,	52,	141,	186,	193,	53,	49,	
137,	191,	132,	156,	187,	65,	49,	252,	134,	65,	
69,	212,	236,	192,	125,	173,	183,	64,	144,	228,	
28,	194,	218,	53,	21,	193,	47,	38,	204,	192,	
133,	96,	19,	65,	211,	186,	200,	65,	70,	239,	
180,	193,	201,	229,	177,	192,	45,	104,	249,	192,	
21,	27,	161,	193,	8,	132,	11,	194,	187,	186,	
58,	193,	148,	212,	133,	193,	173,	94,	75,	64,	
165,	252,	88,	193,	55,	204,	231,	193,	212,	81,	
202,	193,	109,	44,	152,	64,	237,	65,	99,	192,	
18,	48,	103,	193,	141,	231,	221,	65,	227,	244,	
139,	193,	206,	72,	229,	192,	131,	169,	125,	192,	
87,	81,	18,	193,	113,	80,	44,	193,	73,	232,	
182,	64,	61,	137,	38,	65,	241,	78,	0,	193,	
189,	21,	126,	193,	94,	72,	179,	63,	251,	80,	
181,	193,	240,	46,	174,	64,	10,	18,	149,	193,	
126,	79,	237,	65,	211,	229,	169,	192,	97,	33,	
108,	64,	59,	160,	169,	193,	73,	152,	0,	193,	
7,	193,	187,	193,	244,	137,	217,	193,	75,	98,	
175,	65,	181,	23,	158,	65,	141,	41,	15,	65,	
203,	174,	67,	193,	46,	240,	131,	64,	189,	125,	
184,	192,	253,	126,	46,	193,	45,	37,	72,	193,	
163,	117,	150,	64,	186,	80,	192,	192,	250,	11,	
135,	65,	157,	120,	51,	65,	177,	175,	150,	193,	
147,	167,	179,	193,	178,	178,	118,	193,	29,	232,	
231,	193,	210,	165,	201,	192
};


/*     End of NN defines, variables and data                      */
/******************************************************************/


/* in/out definitions */
const int TCSLED = 3; // TCS's white led   
const int VIBRATION = 5;    //vibration motor
const int IRED = 7; // int.0
const int PIN_SERVO = 9;  //servo sorting
const int M_RUN = 11; //motor speed control (PWM)
const int M_DIR = 12; //motor direction control
const int SWITCH = A0;

Adafruit_TCS34725 tcs = Adafruit_TCS34725(TCS34725_INTEGRATIONTIME_24MS, TCS34725_GAIN_16X);
Servo servo;

uint16_t red, green, blue, cclear;

bool calibrate = true;

uint8_t motorSpeed = 128;   //default motor PWM value

volatile uint16_t watchDogCnt = 0;
const int wdLimit = 3000;   //watchdog limit to raise WatchDogError flag
volatile bool watchDogError = false;

int8_t vessels[N_VESSELS]; //array for assigned color idx of currently sorted beads

uint16_t beadsCnt[N_LABELS];    //+1 for  non-confident beads

uint8_t nonesCnt = 0;   //counts number of nones (chambers with no or black beads - to make vibration)

uint16_t beadToTCSdelayOffset = (uint16_t)(TARGET_PERIOD_MS / 3);   //by design: (15/45deg shift of color sensor location)

/*function declarations*/

/// @brief Identifies beads color using neural network 
/// @param[in]  red     The value of red channel of current bead 
/// @param[in]  green   The value of green channel of current bead
/// @param[in]  blue    The value of blue channel of current bead
/// @param[out] *p      a pointer to variable to store probability value (0-100)
/// @return     icol    color index according to labels definitions 
/// @remark  It refers to additional parameters in program space where NN defines, variables and data provided by Octave are stored
int8_t nn_identify_color(uint16_t red, uint16_t green, uint16_t blue, int8_t *p);

/// @brief Waits for next bead. It checks state of signal from phototransistor
/// @param[in]  offs    a delay to wait until exit from this function 
/// @return     none
void wait_for_next_bead(uint16_t offs);

/// @brief Calibrates angles of drainpipe driven by a servo
/// @param      none
/// @return     err     0 if OK, 1 if error
int8_t calibrate_servo(void);

/// @brief Calibrates motor speed
/// @param      none
/// @return     err     0 if OK, 1 if error
int8_t calibrate_motor_speed(void);

/// @brief Calculates an average point by point (number of values to be averaged is not required)
/// @param[in]  n       current sample number 
/// @param[in]  prevAve previous average of (n-1) values  
/// @param[in]  val     n-th value to be averaged  
/// @return     ave     An average value of n samples
/// @remarks    Use: Ave =  point_average(n, Ave, val)
float point_average(uint16_t n, float prevAve, float val);

// Interrupt is called once a millisecond,
SIGNAL(TIMER0_COMPA_vect)
{

    watchDogCnt++;
    if (watchDogCnt > wdLimit)
    {
        watchDogError = true;
    }

}

void clear_watchdog(void)
{
    watchDogCnt = 0;
    watchDogError = false;

}

void setup() {
    for (byte i = 0; i < 13; i++)
        pinMode(i, INPUT); // not used; power save

    pinMode(A0, INPUT); // switch
    pinMode(A1, INPUT); // not used; power save
    pinMode(A2, INPUT); // not used; power save
    pinMode(A3, INPUT); // not used; power save
    //pinMode(A4, INPUT); // I2C
    //pinMode(A5, INPUT); // I2C
    pinMode(A6, INPUT); // not used; power save
    pinMode(A7, INPUT); // not used; power save

    Serial.begin(115200);
    Serial.println(F("Sorter is starting...."));
    pinMode(SWITCH, INPUT_PULLUP); // switch
    pinMode(M_DIR, OUTPUT);            // direction
    pinMode(M_RUN, OUTPUT);    // PWM B T2 controlling motor
    pinMode(VIBRATION, OUTPUT);
    pinMode(LED_BUILTIN, OUTPUT);
    pinMode(IRED, INPUT);
    pinMode(TCSLED, OUTPUT);

    for (int i=0; i<6; i++)
    {
        digitalWrite(TCSLED, LOW);
        delay(250);
        digitalWrite(TCSLED, HIGH);
        delay(250);
    }
    digitalWrite(VIBRATION, HIGH);
    delay(250);
    digitalWrite(VIBRATION, LOW);

    TCCR2A = _BV(COM2A1) | _BV(WGM21) | _BV(WGM20);
    TCCR2B = _BV(CS20); //start, no-presc.

    //Analog comparator configuration
    //  pinMode(A0, INPUT);
    //  pinMode(A1, INPUT);
    //  DIDR1 = (1<<AIN1D) | (1<< AIN0D);
    //  ADCSRA &= ~(1<<ADEN); // ADC disabled
    //  ACSR = ((1<<ACIE)  | (1<<ACIS0) | (1<<ACIS1)); // ACOMP Interrupt enabled & bandgap ref | (1<<ACBG) used, Rising Edge

    servo.attach(PIN_SERVO, 400, 2700);
    servo.write(0);

    digitalWrite(M_DIR, M_FORWARD);
    motor_speed(motorSpeed);
    // Timer0 is already used for millis() - we'll just interrupt somewhere
    // in the middle and call the "Compare A" function below
    OCR0A = 0x2F;
    TIMSK0 |= _BV(OCIE0A);

    digitalWrite(LED_BUILTIN, LOW);

    for (int i=0; i<N_VESSELS; i++)
        vessels[i] = -1;    //-1, vessel not assigned

    /* Perform calibration if switch is pressed during restarting system*/
    uint8_t temp;
    EEPROM.get(EEADR_CALIBR_DONE, temp);

    if (digitalRead(A0) == LOW || temp != CALIBRATION_DONE_CHAR)
    {
        Serial.println(F("*** CALIBRATION PROCEDURE ***"));
        motor_speed(0);
        int8_t err = 0;
        err = calibrate_servo();
        if (!err) 
            err = calibrate_motor_speed();

        if (err == 0)
        {
            Serial.print(F("\r\nStoring calibration char..."));
            EEPROM.put(EEADR_CALIBR_DONE, CALIBRATION_DONE_CHAR);  //flag of performed calibration
            uint8_t temp;
            EEPROM.get(EEADR_CALIBR_DONE, temp);
            if (temp == CALIBRATION_DONE_CHAR)
            {
                Serial.println(F("OK"));
            }
            else
            {
                Serial.println(F("Failed"));
            }
        }

        Serial.println(F("*** END OF CALIBRATION ***"));
    }

    /* beads counters (just for statistics)*/
    for (int i = 0; i < N_LABELS; i++)
        beadsCnt[i] = 0;
    
    interrupts();
}

void loop() 
{
    motor_speed(motorSpeed);
    
    while(1)
    {
        clear_watchdog();
        wait_for_next_bead(beadToTCSdelayOffset);
    
        if(watchDogError)
        {
            digitalWrite(M_DIR, M_BACKWARD);
            digitalWrite(VIBRATION, HIGH);
            motor_speed(motorSpeed);
            delay(930+465); //1.5x period
            digitalWrite(M_DIR, M_FORWARD);
            digitalWrite(VIBRATION, LOW);
        }
        else
            break;
    }

    uint16_t cmax = 0;
    uint16_t rmax = 0;
    uint16_t gmax = 0;
    uint16_t bmax = 0;

    //scan the bead's color and select max value
    for(uint8_t n = 0; n < 5; n++)
    {   
        tcs.getRawData(&red, &green, &blue, &cclear);

        if (cclear > cmax)
        {
            cmax = cclear;
            rmax = red;
            gmax = green;
            bmax = blue;
        }
    }
    // identify color
    int8_t confidence;
    int8_t icol = nn_identify_color(rmax, gmax, bmax, &confidence);
    
    beadsCnt[icol] = beadsCnt[icol] + 1;
    int8_t vidx = get_vessel_idx(icol);

    uint8_t angle;
    EEPROM.get(EEADR_SERVO_BASE + vidx, angle);

    motor_speed(0);

    if (icol == 0)
        nonesCnt++;     //count nones
    else
        nonesCnt = 0;   //reset counter otherwise, VIBRATION not required, there are some beads
    
    if (nonesCnt > 5)
    {
        digitalWrite(VIBRATION, HIGH);
        delay(1000);
        digitalWrite(VIBRATION, LOW);
        nonesCnt = 0;
    }

    uint8_t switchState = digitalRead(A0);

    delay(700);     //wait for bead to drain
    
    if (switchState)
    {
        servo.write(angle);
    }
    else
    {
        //stop gear motor
        motor_speed(0);

        //turn off TCS led (power save) 
        digitalWrite(TCSLED, LOW);

        //park drain to the thrash vessel
        EEPROM.get(EEADR_SERVO_BASE + THRASH_VESSEL_IDX, angle);
        servo.write(angle);

        //wait for releasing switch
        while(digitalRead(A0) == LOW);
        delay(500);

        //detach servo (power save)
        servo.detach();

        /*display statistc*/
        Serial.println(F("Statistics:"));
        long beadsCntTotal = 0;
        for (int i=0; i<N_LABELS; i++)
        {
            //pgm_read_word(&(labelTable[icol]))
            Serial.print(F("Label "));
            Serial.print(i);
            Serial.print(" [");
            char labelName[15];
            strcpy_P(labelName, (char *)pgm_read_word(&(labelTable[i])));
            Serial.print(labelName);
            Serial.print("]\t: ");
            Serial.println(beadsCnt[i]);
            if (i) 
                beadsCntTotal += beadsCnt[i];    //sum all sorted beads so far except NONE/BLACK beads.
        }
        Serial.print(F("TOTAL BEADS (excl. NONES): "));
        Serial.println(beadsCntTotal);

        /*stop until switch_toggles again*/
        while(digitalRead(A0) == HIGH)
        {
            /*blink onboard led just to show it is alive*/
            digitalWrite(LED_BUILTIN, HIGH);
            delay(100);
            digitalWrite(LED_BUILTIN, LOW);
            delay(100);
        }

        long time = millis();
        while(digitalRead(A0) == LOW)
        {
            if ( (millis() - time) > 3000)
            {
                /* reset statistics */
                for (int i = 0; i < N_LABELS; i++)
                    beadsCnt[i] = 0;
                /* reset vessels assignment*/
                for (int i = 0; i < N_VESSELS; i++)
                    vessels[i] = -1;    //-1, vessel not assigned

                Serial.println(F("Sorting reset."));
                Serial.println(F("New bead colors will be sorted."));
                break;
            }
        }

        /*check if vessels assignment hasn't been reset*/
        /*which means sorting is continued but not reset*/
        if (vessels[1] != -1)
        {
            Serial.println(F("Sorting is continued..."));
        }

        digitalWrite(LED_BUILTIN, LOW);  //led off
        servo.attach(PIN_SERVO, 400, 2700);
        digitalWrite(TCSLED, HIGH);
        motor_speed(motorSpeed);
    }


    //digitalWrite(LED_BUILTIN, LOW);  //led off
}


int8_t get_vessel_idx(int8_t icol)
{
    if (icol)
    {
        for (uint8_t i = 0; i < N_VESSELS; i++)
        {
            if ( i == THRASH_VESSEL_IDX)
                continue;

            if (vessels[i] == icol) 
                return i;

            if (vessels[i] == -1)
            {
                vessels[i] = icol;
                return i;
            }
            
        }
    }
    /* assign thrash for NONES, BLACKS, and UNCONFIDENT*/
    return THRASH_VESSEL_IDX;
}

void wait_for_next_bead(uint16_t offs)
{
    int time = millis();

    /*wait till the end of current hole*/
    delay(20);

    while (!digitalRead(IRED) && !watchDogError);

    delay(20);

    while (digitalRead(IRED) && !watchDogError);

    delay(20);

    while (!digitalRead(IRED) && !watchDogError);
    
    if (offs) delay(offs);

}

float sigmoid(float z)
{
    return 1 / (1 + exp(-z));
}

/*define useful global values*/
/*calculate average value of actual beads sorter's white bead RGB values*/

int8_t nn_identify_color(uint16_t red, uint16_t green, uint16_t blue, int8_t *p)
{
    uint16_t nInputLayer, nLabels, nHiddenLayer;
    uint32_t time = millis();

    //calculate average over RGB of averaged white beads from NN training data
    float NNwav = 0;      //average value of white RBG of your beads sorter
    for (uint8_t i = 0; i<3; i++)
    {
        NNwav += pgm_read_float_near(NN_wav + i);
    }
    NNwav /= 3;

    /* Calculate white balance compensation coefficients*/
    /* - it includes brightness normalization to the training average */
    float wbc[3];
    for (uint8_t i = 0; i<3; i++)
    {
        wbc[i] = NNwav / pgm_read_float_near(BS_wav + i);
    }

    /*pre-processing of input values*/

    __nn_X[0] = 1; // bias term
    
    /*perform white balace (wbc) and normalize RGB signals*/
    /* X = (RGB * wbc - mu) / sigma */

    __nn_X[1] = (float)red   * wbc[0] - pgm_read_float_near(NN_mu);
    __nn_X[2] = (float)green * wbc[1] - pgm_read_float_near(NN_mu + 1);
    __nn_X[3] = (float)blue  * wbc[2] - pgm_read_float_near(NN_mu + 2);
    __nn_X[1] /= pgm_read_float_near(NN_sigma);
    __nn_X[2] /= pgm_read_float_near(NN_sigma + 1);
    __nn_X[3] /= pgm_read_float_near(NN_sigma + 2);

    //theta index
    uint16_t theta_offs = 0;

    //temporary variable 
    float h = 0;

    // Serial.println(F("a2="));
    __nn_a2[0] = 1; // bias term
    for (uint8_t i = 0; i < N_HIDDEN_LAYERS; i++)
    {
        h = 0;
        for (uint8_t j = 0; j < (N_INPUT_LAYERS + 1); j++)
        {
            h += __nn_X[j] * pgm_read_float_near(NN_thetas + theta_offs * 4);
            theta_offs++;
        }
        __nn_a2[i + 1] = sigmoid(h);
        // Serial.println(a2[i + 1], 12);
    }

    // calculate final result
    uint8_t idxMax = 0;
    float a3 = 0;
    float a3max = 0;
    float th;
    for (uint8_t i = 0; i < N_LABELS; i++)
    {
        h = 0;
        for (uint8_t j = 0; j < (N_HIDDEN_LAYERS + 1); j++)
        {
            th = pgm_read_float_near(NN_thetas + theta_offs * 4);
            h += __nn_a2[j] * th;
            theta_offs++;
        }

        a3 = sigmoid(h);

        if (a3 > a3max)
        {
            a3max = a3;
            idxMax = i;
        }
    }
    *p = (int8_t)(a3max * 100);

    //time = millis() - time;

    char labelName[15];
    strcpy_P(labelName, (char *)pgm_read_word(&(labelTable[idxMax])));

//    Serial.print(time);
//    Serial.print("\t");
    Serial.print(red);
    Serial.print(",");
    Serial.print(green);
    Serial.print(",");
    Serial.print(blue);
    Serial.print(",");
    Serial.print(red + green + blue);
    Serial.print(",");
    Serial.print(idxMax);
    Serial.print(",");
    Serial.print(*p);
    Serial.print(",");
    Serial.println(labelName);

    return idxMax;
}

int8_t calibrate_servo(void)
{
  char inChar;
  uint8_t servoValue = 90;
  int8_t sn;
  int8_t err = 0;

  for (sn = 0; sn <N_VESSELS; sn++)
  {

    Serial.println();
    Serial.print(F("Calibrate vessel "));
    Serial.print(sn);
    Serial.println(F("[use a=left d=right. ENTER=finish]"));

    servo.write(servoValue);
    delay(500);

    Serial.flush();
    while (1)
    {
      while (!Serial.available())
        ;
      inChar = Serial.read();

      if (inChar == 'a' || inChar == 'A')
      {
        if (servoValue > 0)
          servoValue--;
        servo.write(servoValue);
        Serial.print(F("Angle="));
        Serial.print(servoValue);
        Serial.write(' ');
        Serial.write('\r');
      }
      else if (inChar == 'd' || inChar == 'D')
      {
        if (servoValue < 180)
          servoValue++;
        servo.write(servoValue);
        Serial.print(F("Angle="));
        Serial.print(servoValue);
        Serial.write(' ');
        Serial.write('\r');
      }
      else if (inChar == '\r' || inChar == '\n')
        break;
      // delay(50);
    }
    Serial.print(F("Angle value = "));
    Serial.println(servoValue);

    //Serial.println(F("  Storing to EEPROM...NOT IMPLEMENTED"));
    
    Serial.print(F("  Storing to EEPROM..."));
    EEPROM.put(EEADR_SERVO_BASE + sn, servoValue);
    uint8_t temp;
    EEPROM.get(EEADR_SERVO_BASE + sn, temp);
    if (temp == servoValue)
    {
      Serial.println(F("OK"));
    }
    else
    {
      Serial.println(F("Failed"));
      err++;
    }
    
  }

  return err;
}

int8_t calibrate_motor_speed(void)
{
    uint16_t r, g, b, c;
    char txt[100];
    uint8_t ave = 21;
    int8_t err = 0;

    motor_speed(motorSpeed);
    uint32_t time = millis();
    uint32_t dt = 0;
    float period = 0;

    Serial.println(F("\r\n\r\nStart calibration of motor speed:"));

    clear_watchdog();

    for(int i = 0; i < ave; i++)
    {
        wait_for_next_bead(0);
        if (watchDogError)
        {
            period = 0;
            break;
        }
        clear_watchdog();

        dt = millis() - time;
        time = millis();
        sprintf(txt, "n=%d dt= %ld ", i, dt);
        Serial.println(txt);
        if (i)
            period = point_average(i, period, (float)dt);
    }
    sprintf(txt, "Ave_dt= %d", (uint16_t)period);
    Serial.println(txt);
    
    /*calibrate motor PWM*/
    float val = (float)motorSpeed * period / TARGET_PERIOD_MS;
    /*modify motorSpeed only if result is reasonable*/
    if (val > 50 && val < 255)
    {
        Serial.print(F("Storing motor PWM to EEPROM..."));
        motorSpeed = (uint8_t)val;
        EEPROM.put(EEADR_MOTOR_PWM, motorSpeed);

        uint8_t temp;
        EEPROM.get(EEADR_MOTOR_PWM, temp);
        if (temp == motorSpeed)
        {
            Serial.println(F("OK"));
        }
        else
        {
            Serial.println(F("Failed"));
            Serial.println(F("Use default value 128"));
            motorSpeed = 128;
        }
    }
    else
    {
        err = 1;
        Serial.println(F("ERR: Calculated motor PWM not reasonable (out of 50..255 range)."));
        Serial.println(F("Try to check: \n\r1) IR led and phototransistor wiring, \n\r2) DC motor wiring and/or RPM value, \n\r3) if main spur gear not blocked"));
    }

    return err;
}

float point_average(uint16_t n, float prevAve, float actVal)
{
    float result;

    if (n == 0)
        return 0;
    
    result = (prevAve * (n - 1) + actVal) / n;

    return result;
}